@@ -31,15 +31,21 @@ namespace clad {
   namespace utils {
     class StmtClone;
   }
-  class FunctionDeclInfo;
+  class DiffRequest;
+  namespace plugin {
+    class CladPlugin;
+    clang::FunctionDecl* ProcessDiffRequest(CladPlugin& P, DiffRequest& request);
+  }
 }
 
 namespace clad {
   /// A pair of FunctionDecl and potential enclosing context, e.g. a function
   // in nested namespaces
   using DeclWithContext = std::pair<clang::FunctionDecl*, clang::Decl*>;
+
+  using DiffParams = llvm::SmallVector<const clang::VarDecl*, 16>;
+
   static clang::SourceLocation noLoc{};
-  class DiffPlan;
   /// The main builder class which then uses either ForwardModeVisitor or
   /// ReverseModeVisitor based on the required mode.
   class DerivativeBuilder {
@@ -49,6 +55,7 @@ namespace clad {
     friend class ReverseModeVisitor;
 
     clang::Sema& m_Sema;
+    plugin::CladPlugin& m_CladPlugin;
     clang::ASTContext& m_Context;
     std::unique_ptr<utils::StmtClone> m_NodeCloner;
     clang::NamespaceDecl* m_BuiltinDerivativesNSD;
@@ -57,9 +64,19 @@ namespace clad {
                             llvm::SmallVectorImpl<clang::Expr*>& CallArgs);
     bool overloadExists(clang::Expr* UnresolvedLookup,
                             llvm::MutableArrayRef<clang::Expr*> ARargs);
-
+    /// Shorthand to issues a warning or error.
+    template <std::size_t N>
+    void diag(clang::DiagnosticsEngine::Level level, // Warning or Error
+              clang::SourceLocation loc,
+              const char (&format)[N],
+              llvm::ArrayRef<llvm::StringRef> args = {}) {
+      unsigned diagID = m_Sema.Diags.getCustomDiagID(level, format);
+      clang::Sema::SemaDiagnosticBuilder stream = m_Sema.Diag(loc, diagID);
+      for (auto arg : args)
+        stream << arg;
+    }
   public:
-    DerivativeBuilder(clang::Sema& S);
+    DerivativeBuilder(clang::Sema& S, plugin::CladPlugin& P);
     ~DerivativeBuilder();
 
     ///\brief Produces the derivative of a given function
@@ -70,7 +87,8 @@ namespace clad {
     ///\returns The differentiated function and potentially created enclosing
     /// context.
     ///
-    DeclWithContext Derive(FunctionDeclInfo& FDI, const DiffPlan & plan);
+    DeclWithContext Derive(const clang::FunctionDecl* FD,
+                           const DiffRequest & request);
   };
 
   /// A base class for all common functionality for visitors
@@ -79,24 +97,38 @@ namespace clad {
     VisitorBase(DerivativeBuilder& builder) :
       m_Builder(builder),
       m_Sema(builder.m_Sema),
+      m_CladPlugin(builder.m_CladPlugin),
       m_Context(builder.m_Context),
       m_CurScope(m_Sema.TUScope),
-      m_DerivativeInFlight(false) {}
+      m_DerivativeInFlight(false),
+      m_Derivative(nullptr),
+      m_Function(nullptr) {}
 
     using Stmts = llvm::SmallVector<clang::Stmt*, 16>;
 
     DerivativeBuilder& m_Builder;
     clang::Sema& m_Sema;
+    plugin::CladPlugin& m_CladPlugin;
     clang::ASTContext& m_Context;
     clang::Scope* m_CurScope;
     bool m_DerivativeInFlight;
     /// The Derivative function that is being generated.
     clang::FunctionDecl* m_Derivative;
     /// The function that is currently differentiated.
-    clang::FunctionDecl* m_Function;
+    const clang::FunctionDecl* m_Function;
+
+    /// Map used to keep track of variable declarations and match them
+    /// with their derivatives.
+    std::unordered_map<const clang::VarDecl*, clang::Expr*> m_Variables;
+    /// Map contains variable declarations replacements. If the original
+    /// function contains a declaration which name collides with something
+    /// already created inside derivative's body, the declaration is replaced
+    /// with a new one.
+    /// See the example inside ForwardModeVisitor::VisitDeclStmt.
+    std::unordered_map<const clang::VarDecl*, clang::VarDecl*> m_DeclReplacements;
     /// A stack of all the blocks where the statements of the gradient function
     /// are stored (e.g., function body, if statement blocks).
-    std::stack<Stmts> m_Blocks;
+    std::vector<Stmts> m_Blocks;
 
     template <typename Range>
     clang::CompoundStmt* MakeCompoundStmt(const Range & Stmts) {
@@ -109,17 +141,17 @@ namespace clad {
 
     /// Get the latest block of code (i.e. place for statements output).
     Stmts& getCurrentBlock() {
-      return m_Blocks.top();
+      return m_Blocks.back();
     }
     /// Create new block.
     Stmts& beginBlock() {
-      m_Blocks.push({});
-      return m_Blocks.top();
+      m_Blocks.push_back({});
+      return m_Blocks.back();
     }
     /// Remove the block from the stack, wrap it in CompoundStmt and return it.
     clang::CompoundStmt* endBlock() {
       auto CS = MakeCompoundStmt(getCurrentBlock());
-      m_Blocks.pop();
+      m_Blocks.pop_back();
       return CS;
     }
 
@@ -128,6 +160,7 @@ namespace clad {
     /// Output a statement to the current block. If Stmt is null or is an unused
     /// expression, it is not output and false is returned. 
     bool addToCurrentBlock(clang::Stmt* S);
+    bool addToBlock(clang::Stmt* S, Stmts& block);
 
     /// Get a current scope.
     clang::Scope* getCurrentScope() {
@@ -179,7 +212,7 @@ namespace clad {
     clang::DeclStmt* BuildDeclStmt(llvm::MutableArrayRef<clang::Decl*> DS);
 
     /// Builds a DeclRefExpr to a given Decl.
-    clang::DeclRefExpr* BuildDeclRef(clang::VarDecl* D);
+    clang::DeclRefExpr* BuildDeclRef(clang::DeclaratorDecl* D);
 
     /// Stores the result of an expression in a temporary variable (of the same
     /// type as is the result of the expression) and returns a reference to it.
@@ -187,26 +220,27 @@ namespace clad {
     /// variable declaration. Otherwise, temporary variable is created only 
     /// if E requires evaluation (e.g. there is no point to store literals or
     /// direct references in intermediate variables)
+    clang::Expr* StoreAndRef(clang::Expr* E, Stmts& block,
+                             llvm::StringRef prefix = "_t",
+                             bool forceDeclCreation = false);
+    /// A shorthand to store directly to the current block.
     clang::Expr* StoreAndRef(clang::Expr* E,
                              llvm::StringRef prefix = "_t",
                              bool forceDeclCreation = false);
     /// An overload allowing to specify the type for the variable.
-    clang::Expr* StoreAndRef(clang::Expr* E,
-                             clang::QualType Type,
+    clang::Expr* StoreAndRef(clang::Expr* E, clang::QualType Type, Stmts& block,
                              llvm::StringRef prefix = "_t",
                              bool forceDeclCreation = false);
-
+    /// A flag for silencing warnings/errors output by diag function.
+    bool silenceDiags = false;
     /// Shorthand to issues a warning or error.
     template <std::size_t N>
     void diag(clang::DiagnosticsEngine::Level level, // Warning or Error
               clang::SourceLocation loc,
               const char (&format)[N],
               llvm::ArrayRef<llvm::StringRef> args = {}) {
-      unsigned diagID
-        = m_Sema.Diags.getCustomDiagID(level, format);
-      clang::Sema::SemaDiagnosticBuilder stream = m_Sema.Diag(loc, diagID);
-      for (auto arg : args)
-        stream << arg;
+      if (!silenceDiags)
+        m_Builder.diag(level, loc, format, args);
     }
 
     /// Conuter used to create unique identifiers for temporaries
@@ -223,8 +257,33 @@ namespace clad {
     clang::Stmt* Clone(const clang::Stmt* S);
     /// A shorthand to simplify cloning of expressions.
     clang::Expr* Clone(const clang::Expr* E);
-  };
+    /// Parses the argument expression for the clad::differentiate/clad::gradient
+    /// call. The argument is used to specify independent parameter(s) for
+    /// differentiation. There are three valid options for the argument expression:
+    ///   1) A string literal, containing comma-separated names of function's parameters,
+    ///      as defined in function's defintion. The function will be differentiated
+    ///      w.r.t. all the specified parameters.
+    ///   2) A numeric literal. The function will be differentiated w.r.t. to the
+    ///      parameter corresponding to literal's value index.
+    ///   3) If no argument is provided, a default argument is used. The function
+    ///      will be differentiated w.r.t. to its every parameter.
+    DiffParams parseDiffArgs(const clang::Expr* diffArgs,
+                             const clang::FunctionDecl* FD);
 
+    /// Get an expression used to zero-initialize given type.
+    /// Returns 0 for scalar types, otherwise {}.
+    clang::Expr* getZeroInit(clang::QualType T);
+
+    /// Split an array subscript expression into a pair of base expr and 
+    /// a vector of all indices.
+    std::pair<const clang::Expr*, llvm::SmallVector<const clang::Expr*, 4>>
+    SplitArraySubscript(const clang::ArraySubscriptExpr* ASE);
+
+    /// Build an array subscript expression with a given base expression and
+    /// a sequence of indices.
+    clang::Expr* BuildArraySubscript(clang::Expr* Base,
+                                     const llvm::SmallVector<clang::Expr*, 4> & IS);
+  };
   /// A class that represents the result of Visit of ForwardModeVisitor.
   /// Stmt() allows to access the original (cloned) Stmt and Stmt_dx() allows
   /// to access its derivative (if exists, otherwise null). If Visit produces
@@ -277,18 +336,9 @@ namespace clad {
     : public clang::ConstStmtVisitor<ForwardModeVisitor, StmtDiff>,
       public VisitorBase {
   private:
-    clang::VarDecl* m_IndependentVar;
-    /// Map used to keep track of variable declarations and match them
-    /// with their derivatives.
-    std::unordered_map<clang::VarDecl*, clang::Expr*> m_Variables;
-    /// Map contains variable declarations replacements. If the original
-    /// function contains a declaration which name collides with something
-    /// already created inside derivative's body, the declaration is replaced
-    /// with a new one.
-    /// See the example inside ForwardModeVisitor::VisitDeclStmt.
-    std::unordered_map<const clang::VarDecl*, clang::VarDecl*> m_DeclReplacements;
-    unsigned m_DerivativeOrder;
-    unsigned m_ArgIndex;
+    const clang::VarDecl* m_IndependentVar = nullptr;
+    unsigned m_DerivativeOrder = ~0;
+    unsigned m_ArgIndex = ~0;
 
   public:
     ForwardModeVisitor(DerivativeBuilder& builder);
@@ -301,25 +351,27 @@ namespace clad {
     ///\returns The differentiated and potentially created enclosing
     /// context.
     ///
-    DeclWithContext Derive(FunctionDeclInfo& FDI, const DiffPlan& plan);
-
-    StmtDiff VisitStmt(const clang::Stmt* S);
-    StmtDiff VisitCompoundStmt(const clang::CompoundStmt* CS);
-    StmtDiff VisitIfStmt(const clang::IfStmt* If);
-    StmtDiff VisitReturnStmt(const clang::ReturnStmt* RS);
-    StmtDiff VisitUnaryOperator(const clang::UnaryOperator* UnOp);
+    DeclWithContext Derive(const clang::FunctionDecl* FD,
+                           const DiffRequest& request);
+    StmtDiff VisitArraySubscriptExpr(const clang::ArraySubscriptExpr* ASE);
     StmtDiff VisitBinaryOperator(const clang::BinaryOperator* BinOp);
+    StmtDiff VisitCallExpr(const clang::CallExpr* CE);
+    StmtDiff VisitCompoundStmt(const clang::CompoundStmt* CS);
+    StmtDiff VisitConditionalOperator(const clang::ConditionalOperator* CO);
     StmtDiff VisitCXXOperatorCallExpr(const clang::CXXOperatorCallExpr* OpCall);
     StmtDiff VisitDeclRefExpr(const clang::DeclRefExpr* DRE);
-    StmtDiff VisitParenExpr(const clang::ParenExpr* PE);
-    StmtDiff VisitMemberExpr(const clang::MemberExpr* ME);
-    StmtDiff VisitIntegerLiteral(const clang::IntegerLiteral* IL);
-    StmtDiff VisitFloatingLiteral(const clang::FloatingLiteral* FL);
-    StmtDiff VisitCallExpr(const clang::CallExpr* CE);
     StmtDiff VisitDeclStmt(const clang::DeclStmt* DS);
-    StmtDiff VisitImplicitCastExpr(const clang::ImplicitCastExpr* ICE);
-    StmtDiff VisitConditionalOperator(const clang::ConditionalOperator* CO);
+    StmtDiff VisitFloatingLiteral(const clang::FloatingLiteral* FL);
     StmtDiff VisitForStmt(const clang::ForStmt* FS);
+    StmtDiff VisitIfStmt(const clang::IfStmt* If);
+    StmtDiff VisitImplicitCastExpr(const clang::ImplicitCastExpr* ICE);
+    StmtDiff VisitInitListExpr(const clang::InitListExpr* ILE);
+    StmtDiff VisitIntegerLiteral(const clang::IntegerLiteral* IL);
+    StmtDiff VisitMemberExpr(const clang::MemberExpr* ME);
+    StmtDiff VisitParenExpr(const clang::ParenExpr* PE);
+    StmtDiff VisitReturnStmt(const clang::ReturnStmt* RS);
+    StmtDiff VisitStmt(const clang::Stmt* S);
+    StmtDiff VisitUnaryOperator(const clang::UnaryOperator* UnOp);
     // Decl is not Stmt, so it cannot be visited directly.
     VarDeclDiff DifferentiateVarDecl(const clang::VarDecl* VD);
   };
@@ -327,21 +379,94 @@ namespace clad {
   /// A visitor for processing the function code in reverse mode.
   /// Used to compute derivatives by clad::gradient.
   class ReverseModeVisitor
-    : public clang::ConstStmtVisitor<ReverseModeVisitor, void>,
+    : public clang::ConstStmtVisitor<ReverseModeVisitor, StmtDiff>,
       public VisitorBase {
   private:
+    llvm::SmallVector<clang::VarDecl*, 16> m_IndependentVars;
     /// Stack is used to pass the arguments (dfdx) to further nodes
     /// in the Visit method.
     std::stack<clang::Expr*> m_Stack;
     clang::Expr* dfdx () {
-        return m_Stack.top ();
+      if (m_Stack.empty()) {
+        auto IntTy = m_Context.IntTy;
+        llvm::APInt zero(m_Context.getIntWidth(IntTy), 0, /*isSigned*/ true); 
+        return clang::IntegerLiteral::Create(m_Context, zero, m_Context.IntTy, noLoc);
+      }
+      return m_Stack.top();
     }
-    void Visit(const clang::Stmt* stmt, clang::Expr* expr) {
+    StmtDiff Visit(const clang::Stmt* stmt, clang::Expr* expr = nullptr) {
+      if (expr)
         m_Stack.push(expr);
-        clang::ConstStmtVisitor<ReverseModeVisitor, void>::Visit(stmt);
+      auto result = clang::ConstStmtVisitor<ReverseModeVisitor, StmtDiff>::Visit(stmt);
+      if (expr)
         m_Stack.pop();
+      return result;
     }
 
+    /// In addition to a sequence of forward-accumulated Stmts (m_Blocks), in 
+    /// the reverse mode we also accumulate Stmts for the reverse pass which
+    /// will be executed on return.
+    std::vector<Stmts> m_Reverse;
+    /// An enum to operate between forward and reverse passes.
+    enum direction { forward, reverse };
+    /// Get the latest block of code (i.e. place for statements output).
+    Stmts& getCurrentBlock(direction d = forward) {
+      if (d == forward)
+        return m_Blocks.back();
+      else
+        return m_Reverse.back();
+    }
+    /// Create new block.
+    Stmts& beginBlock(direction d = forward) {
+      if (d == forward)
+        m_Blocks.push_back({});
+      else
+        m_Reverse.push_back({});
+      return getCurrentBlock(d);
+    }
+    /// Remove the block from the stack, wrap it in CompoundStmt and return it.
+    clang::CompoundStmt* endBlock(direction d = forward) {
+      if (d == forward) {
+        auto CS = MakeCompoundStmt(getCurrentBlock(forward));
+        m_Blocks.pop_back();
+        return CS;
+      } else {
+        auto R = getCurrentBlock(reverse);
+        std::reverse(std::begin(R), std::end(R));
+        auto CS = MakeCompoundStmt(R);
+        m_Reverse.pop_back();
+        return CS;
+      }
+    }
+    /// Output a statement to the current block. If Stmt is null or is an unused
+    /// expression, it is not output and false is returned. 
+    bool addToCurrentBlock(clang::Stmt* S, direction d = forward) {
+      return addToBlock(S, getCurrentBlock(d));
+    }
+
+    /// Stores the result of an expression in a temporary variable (of the same
+    /// type as is the result of the expression) and returns a reference to it.
+    /// If force decl creation is true, this will allways create a temporary
+    /// variable declaration. Otherwise, temporary variable is created only 
+    /// if E requires evaluation (e.g. there is no point to store literals or
+    /// direct references in intermediate variables)
+    clang::Expr* StoreAndRef(clang::Expr* E, direction d = forward,
+                             llvm::StringRef prefix = "_t",
+                             bool forceDeclCreation = false) {
+      return VisitorBase::StoreAndRef(E, getCurrentBlock(d), prefix, forceDeclCreation);
+    }
+    
+    /// An overload allowing to specify the type for the variable.
+    clang::Expr* StoreAndRef(clang::Expr* E, clang::QualType Type,
+                             direction d = forward,
+                             llvm::StringRef prefix = "_t",
+                             bool forceDeclCreation = false) {
+      return VisitorBase::StoreAndRef(E, Type, getCurrentBlock(d), prefix, forceDeclCreation);
+    }
+    /// Outputs the Stmts in the reverse pass to current block, intended to be
+    /// executed once return is reached.
+    void outputReverse();
+
     //// A reference to the output parameter of the gradient function.
     clang::Expr* m_Result;
 
@@ -356,22 +481,39 @@ namespace clad {
     ///\returns The gradient of the function and potentially created enclosing
     /// context.
     ///
-    DeclWithContext Derive(FunctionDeclInfo& FDI, const DiffPlan& plan);
-    void VisitCompoundStmt(const clang::CompoundStmt* CS);
-    void VisitIfStmt(const clang::IfStmt* If);
-    void VisitReturnStmt(const clang::ReturnStmt* RS);
-    void VisitUnaryOperator(const clang::UnaryOperator* UnOp);
-    void VisitBinaryOperator(const clang::BinaryOperator* BinOp);
-    void VisitDeclStmt(const clang::DeclStmt* DS);
-    void VisitMemberExpr(const clang::MemberExpr* ME);
-    void VisitArraySubscriptExpr(const clang::ArraySubscriptExpr* ASE);
-    void VisitDeclRefExpr(const clang::DeclRefExpr* DRE);
-    void VisitParenExpr(const clang::ParenExpr* PE);
-    void VisitIntegerLiteral(const clang::IntegerLiteral* IL);
-    void VisitFloatingLiteral(const clang::FloatingLiteral* FL);
-    void VisitCallExpr(const clang::CallExpr* CE);
-    void VisitImplicitCastExpr(const clang::ImplicitCastExpr* ICE);
-    void VisitConditionalOperator(const clang::ConditionalOperator* CO);
+    /// We name the gradient of f as 'f_grad'.
+    /// If the gradient of the same function is requested several times
+    /// with different parameters, but same parameter types, every such request
+    /// will create f_grad function with the same signature, which will be
+    /// ambiguous. E.g.
+    ///   double f(double x, double y, double z) { ... }
+    ///   clad::gradient(f, "x, y");
+    ///   clad::gradient(f, "x, z");
+    /// will create 2 definitions for f_grad with the same signature.
+    ///
+    /// Improved naming scheme is required. Hence, we append the indices to of the
+    /// requested parameters to 'f_grad', i.e. in the previous example "x, y" will
+    /// give 'f_grad_0_1' and "x, z" will give 'f_grad_0_2'.
+    DeclWithContext Derive(const clang::FunctionDecl* FD,
+                           const DiffRequest& request);
+    StmtDiff VisitArraySubscriptExpr(const clang::ArraySubscriptExpr* ASE);
+    StmtDiff VisitBinaryOperator(const clang::BinaryOperator* BinOp);
+    StmtDiff VisitCallExpr(const clang::CallExpr* CE);
+    StmtDiff VisitCompoundStmt(const clang::CompoundStmt* CS);
+    StmtDiff VisitConditionalOperator(const clang::ConditionalOperator* CO);
+    StmtDiff VisitDeclRefExpr(const clang::DeclRefExpr* DRE);
+    StmtDiff VisitDeclStmt(const clang::DeclStmt* DS);
+    StmtDiff VisitFloatingLiteral(const clang::FloatingLiteral* FL);
+    StmtDiff VisitIfStmt(const clang::IfStmt* If);
+    StmtDiff VisitImplicitCastExpr(const clang::ImplicitCastExpr* ICE);
+    StmtDiff VisitInitListExpr(const clang::InitListExpr* ILE);
+    StmtDiff VisitIntegerLiteral(const clang::IntegerLiteral* IL);
+    StmtDiff VisitMemberExpr(const clang::MemberExpr* ME);
+    StmtDiff VisitParenExpr(const clang::ParenExpr* PE);
+    StmtDiff VisitReturnStmt(const clang::ReturnStmt* RS);
+    StmtDiff VisitUnaryOperator(const clang::UnaryOperator* UnOp);
+    // Decl is not Stmt, so it cannot be visited directly.
+    VarDeclDiff DifferentiateVarDecl(const clang::VarDecl* VD);
   };
 } // end namespace clad
 
